@@ -53,6 +53,10 @@ struct _PtWaveviewerPrivate {
 	gboolean  fixed_cursor;
 	gboolean  show_ruler;
 	gboolean  has_selection;
+	gboolean  zoom;
+
+	gint64    zoom_time;
+	gint      zoom_pos;
 
 	/* Selections, in milliseconds */
 	gint64    sel_start;
@@ -84,6 +88,7 @@ enum
 	PROP_HAS_SELECTION,
 	PROP_SELECTION_START,
 	PROP_SELECTION_END,
+	PROP_ZOOM,
 	N_PROPERTIES
 };
 
@@ -91,6 +96,8 @@ enum {
 	CURSOR_CHANGED,
 	SELECTION_CHANGED,
 	PLAY_TOGGLED,
+	ZOOM_IN,
+	ZOOM_OUT,
 	LAST_SIGNAL
 };
 
@@ -504,6 +511,36 @@ pt_waveviewer_button_press_event (GtkWidget      *widget,
 	return FALSE;
 }
 
+static gboolean
+pt_waveviewer_scroll_event (GtkWidget      *widget,
+			    GdkEventScroll *event)
+{
+	PtWaveviewer *self = PT_WAVEVIEWER (widget);
+
+	/* No modifier pressed: scrolling back and forth */
+	if (!(event->state & ALL_ACCELS_MASK)) {
+		gtk_propagate_event
+			(gtk_scrolled_window_get_hscrollbar (GTK_SCROLLED_WINDOW (widget)),
+			(GdkEvent*)event);
+		return TRUE;
+	}
+
+	/* Only Control pressed: zoom in or out */
+	if ((event->state & ALL_ACCELS_MASK) == GDK_CONTROL_MASK) {
+		if (event->delta_y < 0 || event->delta_x < 0) {
+			g_signal_emit_by_name (self, "zoom-out");
+			return TRUE;
+		}
+		if (event->delta_y > 0 || event->delta_x > 0) {
+			g_signal_emit_by_name (self, "zoom-in");
+			return TRUE;
+		}
+
+	}
+
+	return FALSE;
+}
+
 static gboolean
 pt_waveviewer_motion_notify_event (GtkWidget      *widget,
 				   GdkEventMotion *event)
@@ -559,7 +596,7 @@ pt_waveviewer_button_release_event (GtkWidget      *widget,
 static void
 update_cached_style_values (PtWaveviewer *self)
 {
-	/* Colors and direction are cached */
+	/* Only direction is cached */
 
 	GtkStyleContext *context;
 	GtkStateFlags    state;
@@ -594,6 +631,31 @@ pt_waveviewer_style_updated (GtkWidget *widget)
 	update_cached_style_values (self);
 }
 
+static void
+pt_waveviewer_direction_changed (GtkWidget *widget,
+				 GtkTextDirection direction)
+{
+	PtWaveviewer *self = PT_WAVEVIEWER (widget);
+	gint widget_width, first_visible, page_width;
+
+	/* This is called after state-flags-changed signal.
+	   The waveform handles repainting itself, here we have to flip
+	   the GtkAdjustment and recalculate the cursor's position */
+	widget_width = gtk_widget_get_allocated_width (self->priv->waveform);
+	first_visible = (gint) gtk_adjustment_get_value (self->priv->adj);
+	page_width = (gint) gtk_adjustment_get_page_size (self->priv->adj);
+	gtk_adjustment_set_value (
+			self->priv->adj,
+			widget_width - first_visible - page_width);
+
+	if (self->priv->peaks)
+		pt_waveviewer_cursor_render (
+				PT_WAVEVIEWER_CURSOR (self->priv->cursor),
+				time_to_pixel (self, self->priv->playback_cursor));
+
+	GTK_WIDGET_CLASS (pt_waveviewer_parent_class)->direction_changed (widget, direction);
+}
+
 static void
 pt_waveviewer_realize (GtkWidget *widget)
 {
@@ -601,16 +663,36 @@ pt_waveviewer_realize (GtkWidget *widget)
 	update_cached_style_values (PT_WAVEVIEWER (widget));
 }
 
+static void
+stop_following_cursor (PtWaveviewer *self)
+{
+	if (self->priv->peaks)
+		pt_waveviewer_set_follow_cursor (self, FALSE);
+}
+
 static gboolean
-scrollbar_cb (GtkWidget      *widget,
-	      GdkEventButton *event,
-	      gpointer        data)
+scrollbar_button_press_event_cb (GtkWidget      *widget,
+				 GdkEventButton *event,
+				 gpointer        data)
 {
 	/* If user clicks on scrollbar don't follow cursor anymore.
 	   Otherwise it would scroll immediately back again. */
 
-	PtWaveviewer *self = PT_WAVEVIEWER (data);
-	pt_waveviewer_set_follow_cursor (self, FALSE);
+	stop_following_cursor (PT_WAVEVIEWER (data));
+
+	/* Propagate signal */
+	return FALSE;
+}
+
+static gboolean
+scrollbar_scroll_event_cb (GtkWidget      *widget,
+			   GdkEventButton *event,
+			   gpointer        data)
+{
+	/* If user scrolls on scrollbar don't follow cursor anymore.
+	   Otherwise it would scroll immediately back again. */
+
+	stop_following_cursor (PT_WAVEVIEWER (data));
 
 	/* Propagate signal */
 	return FALSE;
@@ -744,24 +826,67 @@ pt_waveviewer_set_follow_cursor (PtWaveviewer *self,
 	}
 }
 
+static gboolean
+cursor_is_visible (PtWaveviewer *self)
+{
+	gint cursor_pos, left, right;
+
+	cursor_pos = time_to_pixel (self, self->priv->playback_cursor);
+	left = (gint) gtk_adjustment_get_value (self->priv->adj);
+	right = left + (gint) gtk_adjustment_get_page_size (self->priv->adj);
+
+	return (left <= cursor_pos && cursor_pos <= right);
+}
+
+static void
+size_allocate_cb (GtkWidget    *widget,
+                  GdkRectangle *allocation,
+                  gpointer      user_data)
+{
+	PtWaveviewer *self = PT_WAVEVIEWER (user_data);
+
+	if (self->priv->zoom) {
+		gtk_adjustment_set_value (
+				self->priv->adj,
+				time_to_pixel (self, self->priv->zoom_time) - self->priv->zoom_pos);
+		self->priv->zoom = FALSE;
+	}
+
+	if (self->priv->peaks) {
+		pt_waveviewer_cursor_render (
+				PT_WAVEVIEWER_CURSOR (self->priv->cursor),
+				time_to_pixel (self, self->priv->playback_cursor));
+	}
+
+	pt_waveviewer_selection_set (
+			PT_WAVEVIEWER_SELECTION (self->priv->selection),
+			time_to_pixel (self, self->priv->sel_start),
+			time_to_pixel (self, self->priv->sel_end));
+}
+
 static void
 make_widget_ready (PtWaveviewer *self,
 		   gboolean      ready)
 {
-	gint widget_width;
+	gint widget_width, left;
 
-	if (!ready) {
-		if (self->priv->peaks) {
-			g_free (self->priv->peaks);
-			self->priv->peaks = NULL;
+	if (!ready && self->priv->zoom) {
+		/* If this is a zoom, we still have the old waveform.
+		   Get an anchor point = zoom_pos, either cursor position or
+		   middle of view. The anchor point is relative to the view.
+		   Get time of that anchor point = zoom_time. */
+
+		left = (gint) gtk_adjustment_get_value (self->priv->adj);
+		if (cursor_is_visible (self)) {
+			self->priv->zoom_time = self->priv->playback_cursor;
+			self->priv->zoom_pos = time_to_pixel (self, self->priv->playback_cursor) - left;
+		} else {
+			self->priv->zoom_pos = (gint) gtk_adjustment_get_page_size (self->priv->adj) / 2;
+			self->priv->zoom_time = pixel_to_time (self, self->priv->zoom_pos + left);
 		}
-		self->priv->peaks_size = 0;
-		self->priv->px_per_sec = 0;
-		self->priv->duration = 0;
-		widget_width = -1;
-		pt_waveviewer_cursor_render (
-				PT_WAVEVIEWER_CURSOR (self->priv->cursor), -1);
+	}
 
+	if (!ready && !self->priv->zoom) {
 		/* Reset previous selections */
 		self->priv->sel_start = self->priv->sel_end = 0;
 		self->priv->has_selection = FALSE;
@@ -774,30 +899,48 @@ make_widget_ready (PtWaveviewer *self,
 		g_object_notify_by_pspec (
 				G_OBJECT (self),
 				obj_properties[PROP_HAS_SELECTION]);
+	}
+
+	if (!ready) {
+		if (self->priv->peaks) {
+			g_free (self->priv->peaks);
+			self->priv->peaks = NULL;
+		}
+		self->priv->peaks_size = 0;
+		self->priv->px_per_sec = 0;
+		self->priv->duration = 0;
 		pt_waveviewer_selection_set (
 				PT_WAVEVIEWER_SELECTION (self->priv->selection),
 				0, 0);
+		pt_waveviewer_cursor_render (
+				PT_WAVEVIEWER_CURSOR (self->priv->cursor),
+				-1);
 	} else {
 		self->priv->duration = calculate_duration (self);
 		widget_width = self->priv->peaks_size / 2;
 		gtk_adjustment_set_upper (self->priv->adj, widget_width);
+		gtk_widget_set_size_request (
+				self->priv->waveform,
+				widget_width,
+				WAVE_MIN_HEIGHT);
 	}
 
+	pt_waveviewer_waveform_set (
+			PT_WAVEVIEWER_WAVEFORM (self->priv->waveform),
+			self->priv->peaks,
+			self->priv->peaks_size);
+
 	pt_waveviewer_ruler_set_ruler (
 			PT_WAVEVIEWER_RULER (self->priv->ruler),
 			self->priv->peaks_size / 2,
 			self->priv->px_per_sec,
 			self->priv->duration);
 
-	gtk_widget_set_size_request (
-			self->priv->waveform,
-			widget_width,
-			WAVE_MIN_HEIGHT);
-
-	pt_waveviewer_waveform_set (
-			PT_WAVEVIEWER_WAVEFORM (self->priv->waveform),
-			self->priv->peaks,
-			self->priv->peaks_size);
+	/* Here we would render the cursor and change the GtkAdjustment so that
+	   we show the intended part of the waveform after zooming.
+	   However, there were some issues because the waveform didn't
+	   allocate the new size in time. This is now done in the callback
+	   above, size_allocate_cb(). */
 }
 
 static gboolean
@@ -926,6 +1069,9 @@ pt_waveviewer_set_property (GObject      *object,
 		gtk_revealer_set_reveal_child (GTK_REVEALER (self->priv->revealer),
 					       self->priv->show_ruler);
 		break;
+	case PROP_ZOOM:
+		self->priv->zoom = g_value_get_boolean (value);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
 		break;
@@ -941,7 +1087,11 @@ pt_waveviewer_constructed (GObject *object)
 	self->priv->adj = gtk_scrolled_window_get_hadjustment (GTK_SCROLLED_WINDOW (self));
 	g_signal_connect (gtk_scrolled_window_get_hscrollbar (GTK_SCROLLED_WINDOW (self)),
 			  "button_press_event",
-			  G_CALLBACK (scrollbar_cb),
+			  G_CALLBACK (scrollbar_button_press_event_cb),
+			  self);
+	g_signal_connect (gtk_scrolled_window_get_hscrollbar (GTK_SCROLLED_WINDOW (self)),
+			  "scroll_event",
+			  G_CALLBACK (scrollbar_scroll_event_cb),
 			  self);
 }
 
@@ -977,6 +1127,8 @@ pt_waveviewer_init (PtWaveviewer *self)
 	self->priv->focus_on_cursor = FALSE;
 	self->priv->sel_start = 0;
 	self->priv->sel_end = 0;
+	self->priv->zoom_time = 0;
+	self->priv->zoom_pos = 0;
 	self->priv->arrows = get_resize_cursor ();
 	self->priv->waveform = pt_waveviewer_waveform_new ();
 	self->priv->focus = pt_waveviewer_focus_new ();
@@ -1027,6 +1179,12 @@ pt_waveviewer_init (PtWaveviewer *self)
 	g_object_unref (css_file);
 	g_object_unref (provider);
 
+	g_signal_connect (
+			GTK_WIDGET (self->priv->waveform),
+			"size-allocate",
+			G_CALLBACK (size_allocate_cb),
+			self);
+
 	/* If overriding these vfuncs something's going wrong, note that focus-in
 	   an focus-out need GdkEventFocus as 2nd parameter in vfunc */
 	g_signal_connect (self, "focus", G_CALLBACK (pt_waveviewer_focus), NULL);
@@ -1047,9 +1205,11 @@ pt_waveviewer_class_init (PtWaveviewerClass *klass)
 
 	widget_class->button_press_event   = pt_waveviewer_button_press_event;
 	widget_class->button_release_event = pt_waveviewer_button_release_event;
+	widget_class->direction_changed    = pt_waveviewer_direction_changed;
 	widget_class->key_press_event      = pt_waveviewer_key_press_event;
 	widget_class->motion_notify_event  = pt_waveviewer_motion_notify_event;
 	widget_class->realize              = pt_waveviewer_realize;
+	widget_class->scroll_event         = pt_waveviewer_scroll_event;
 	widget_class->state_flags_changed  = pt_waveviewer_state_flags_changed;
 	widget_class->style_updated        = pt_waveviewer_style_updated;
 
@@ -1107,6 +1267,40 @@ pt_waveviewer_class_init (PtWaveviewerClass *klass)
 		      G_TYPE_NONE,
 		      0);
 
+	/**
+	* PtWaveviewer::zoom-in:
+	* @viewer: the waveviewer emitting the signal
+	*
+	* Signals that the user requested to zoom into the waveform.
+	*/
+	signals[ZOOM_IN] =
+	g_signal_new ("zoom-in",
+		      PT_TYPE_WAVEVIEWER,
+		      G_SIGNAL_RUN_FIRST,
+		      0,
+		      NULL,
+		      NULL,
+		      g_cclosure_marshal_VOID__VOID,
+		      G_TYPE_NONE,
+		      0);
+
+	/**
+	* PtWaveviewer::zoom-out:
+	* @viewer: the waveviewer emitting the signal
+	*
+	* Signals that the user requested to zoom out of the waveform.
+	*/
+	signals[ZOOM_OUT] =
+	g_signal_new ("zoom-out",
+		      PT_TYPE_WAVEVIEWER,
+		      G_SIGNAL_RUN_FIRST,
+		      0,
+		      NULL,
+		      NULL,
+		      g_cclosure_marshal_VOID__VOID,
+		      G_TYPE_NONE,
+		      0);
+
 	/**
 	* PtWaveviewer:playback-cursor:
 	*
@@ -1219,6 +1413,24 @@ pt_waveviewer_class_init (PtWaveviewerClass *klass)
 			0,
 			G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
+	/**
+	* PtWaveviewer:zoom:
+	*
+	* Set this to TRUE to indicate that the next wave data will be the same
+	* waveform but at a different zoom level. PtWaveviewer needs this to
+	* decide which part of the waveform should be shown.
+	* – Note: This will be probably removed with the next API break in
+	* version 1.6 and integrated into pt_waveviewer_set_wave().
+	*/
+
+	obj_properties[PROP_ZOOM] =
+	g_param_spec_boolean (
+			"zoom",
+			"Zoom",
+			"Indicates that next wave data will zoom",
+			FALSE,
+			G_PARAM_WRITABLE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS);
+
 	g_object_class_install_properties (
 			gobject_class,
 			N_PROPERTIES,
